@@ -8,9 +8,11 @@
 #   The GNU Lesser General Public License, Version 2.1, February 1999
 #
 package MooseX::AttributeShortcuts;
-{
-  $MooseX::AttributeShortcuts::VERSION = '0.018'; # TRIAL
+BEGIN {
+  $MooseX::AttributeShortcuts::AUTHORITY = 'cpan:RSRCHBOY';
 }
+# git description: 0.024-8-g8de35d9
+$MooseX::AttributeShortcuts::VERSION = '0.025'; # TRIAL
 
 # ABSTRACT: Shorthand for common attribute options
 
@@ -24,21 +26,29 @@ use Moose::Exporter;
 use Moose::Util::MetaRole;
 use Moose::Util::TypeConstraints;
 
-# debug...
-#use Smart::Comments;
-
 {
     package MooseX::AttributeShortcuts::Trait::Attribute;
-{
-  $MooseX::AttributeShortcuts::Trait::Attribute::VERSION = '0.018'; # TRIAL
+BEGIN {
+  $MooseX::AttributeShortcuts::Trait::Attribute::AUTHORITY = 'cpan:RSRCHBOY';
 }
+# git description: 0.024-8-g8de35d9
+$MooseX::AttributeShortcuts::Trait::Attribute::VERSION = '0.025'; # TRIAL
     use namespace::autoclean;
     use MooseX::Role::Parameterized;
-    use Moose::Util::TypeConstraints qw{ class_type role_type enum };
+    use Moose::Util::TypeConstraints  ':all';
     use MooseX::Types::Moose          ':all';
     use MooseX::Types::Common::String ':all';
 
-    sub _find_or_create_isa_type_constraint { goto \&Moose::Util::TypeConstraints::find_or_create_isa_type_constraint }
+    use autobox::Core;
+    use autobox::Junctions;
+    use List::AllUtils 'any';
+
+    use Package::DeprecationManager -deprecations => {
+        'undocumented-isa-constraints' => '0.23',
+        'hashref-given-to-coerce'      => '0.24',
+    };
+
+    sub _acquire_isa_tc { goto \&Moose::Util::TypeConstraints::find_or_create_isa_type_constraint }
 
     parameter writer_prefix  => (isa => NonEmptySimpleStr, default => '_set_');
     parameter builder_prefix => (isa => NonEmptySimpleStr, default => '_build_');
@@ -94,12 +104,14 @@ use Moose::Util::TypeConstraints;
 
             if ($options->{is}) {
 
+                # handle: is => 'rwp'
                 if ($options->{is} eq 'rwp') {
 
                     $options->{is}     = 'ro';
                     $options->{writer} = "$wprefix$name";
                 }
 
+                # handle: is => 'lazy'
                 if ($options->{is} eq 'lazy') {
 
                     $options->{is}       = 'ro';
@@ -109,24 +121,40 @@ use Moose::Util::TypeConstraints;
                 }
             }
 
-            # TODO isa_class - anon class_type generation
-            # TODO isa_role  - anon role_type generation
-            # TODO isa_enum  - anon enum generation
-            # TODO coerce_via - anon coercion (type -> anon subtype+coercion
+            if (any { $options->keys->any eq $_ } (qw{ isa_class isa_role isa_enum })) {
 
-            # XXX we also ignore conflicts here -- last in wins
-            #confess q{conflict 'isa' and 'isa_class' or 'isa_role'}
-                #if $_has->('isa')
+                # (more than) fair warning...
+                deprecated(
+                    feature => 'undocumented-isa-constraints',
+                    message => 'Naughty! isa_class, isa_role, and isa_enum will be removed on or after 01 July 2014!',
+                );
 
-            # XXX undocumented -- not sure this is a great idea
-            $options->{isa} = class_type(delete $options->{isa_class})
-                if $_has->('isa_class');
-            $options->{isa} = role_type(delete $options->{isa_role})
-                if $_has->('isa_role');
-            $options->{isa} = enum(delete $options->{isa_enum})
-                if $_has->('isa_enum');
+                # XXX undocumented -- not sure this is a great idea
+                $options->{isa} = class_type(delete $options->{isa_class})
+                    if $_has->('isa_class');
+                $options->{isa} = role_type(delete $options->{isa_role})
+                    if $_has->('isa_role');
+                $options->{isa} = enum(delete $options->{isa_enum})
+                    if $_has->('isa_enum');
+            }
+
+            # aka: isa => class_type(...)
+            if ($_has->('isa_instance_of')) {
+
+                if ($_has->('isa')) {
+
+                    $class->throw_error(
+                        q{Cannot use 'isa_instance_of' and 'isa' together for attribute }
+                        . $_opt->('definition_context')->{package} . '::' . $name
+                    );
+                }
+
+                $options->{isa} = class_type(delete $options->{isa_instance_of});
+            }
 
             ### the pretty business of on-the-fly subtyping...
+            my $our_type;
+
             if ($_has->('constraint')) {
 
                 # check for errors...
@@ -140,23 +168,59 @@ use Moose::Util::TypeConstraints;
                     if $_ref->('constraint') eq 'CODE';
 
                 # stash our original option away and construct our new one
-                my $isa = $options->{original_isa} = $_opt->('isa');
-                my $isa_tc = _find_or_create_isa_type_constraint($isa);
-                my $new_tc = $isa_tc->create_child_type(@opts);
+                my $isa     = $options->{original_isa} = $_opt->('isa');
+                $our_type ||= _acquire_isa_tc($isa)->create_child_type(@opts);
+            }
 
-                if ($isa_tc->has_coercion && $_opt->('coerce') eq "1") {
+            # "fix" the case of the hashref....  *sigh*
+            if ($_ref->('coerce') eq 'HASH') {
+
+                deprecated(
+                    feature => 'hashref-given-to-coerce',
+                    message => 'Passing a hashref to coerce is unsafe, and will be removed on or after 01 Jan 2015',
+                );
+
+                $options->{coerce} = [ %{ $options->{coerce} } ];
+            }
+
+            if ($_ref->('coerce') eq 'ARRAY') {
+
+                ### must be type => sub { ... } pairs...
+                my @coercions = @{ $_opt->('coerce') };
+                confess 'You must specify an "isa" when declaring "coercion"'
+                    unless $_has->('isa');
+                confess 'coercion array must be in pairs!'
+                    if @coercions % 2;
+                confess 'must define at least one coercion pair!'
+                    unless @coercions > 0;
+
+                my $our_coercion = Moose::Meta::TypeCoercion->new;
+                $our_type ||= _acquire_isa_tc($_opt->('isa'))->create_child_type;
+
+                $our_coercion->add_type_coercions(@coercions);
+                $our_type->coercion($our_coercion);
+                $options->{coerce} = 1;
+            }
+
+            if ($our_type && !$our_type->has_coercion) {
+
+                my $isa_type = _acquire_isa_tc($_opt->('isa'));
+
+                if ($isa_type->has_coercion && !$_ref->('coerce') && $_opt->('coerce') eq "1") {
 
                     # create our coercion as a copy of the parent
-                    $new_tc->coercion(Moose::Meta::TypeCoercion->new(
-                        type_constraint   => $new_tc,
-                        type_coercion_map => $isa_tc->coercion->type_coercion_map,
+                    $our_type->coercion(Moose::Meta::TypeCoercion->new(
+                        type_constraint   => $our_type,
+                        type_coercion_map => [ @{ $isa_type->coercion->type_coercion_map } ],
                     ));
                 }
 
-                # fin constraint mucking....
-                $options->{isa} = $new_tc;
             }
 
+            # fin constraint mucking....
+            do { $options->{original_isa} = $_opt->('isa'); $options->{isa} = $our_type }
+                if $our_type;
+
             if ($options->{lazy_build} && $options->{lazy_build} eq 'private') {
 
                 $options->{lazy_build} = 1;
@@ -236,6 +300,36 @@ use Moose::Util::TypeConstraints;
             $class->add_method($self->builder => $self->anon_builder);
             return;
         };
+
+        method mi               => sub { shift->associated_class->get_meta_instance                     };
+        method weaken_value     => sub { $_[0]->mi->weaken_slot_value($_[1] => $_) for $_[0]->slots     };
+        method strengthen_value => sub { $_[0]->mi->strengthen_slot_value($_[1] => $_) for $_[0]->slots };
+
+        # NOTE: remove_delegation() will also automagically remove any custom
+        # accessors we create here
+
+        around _make_delegation_method => sub {
+            my ($orig, $self) = (shift, shift);
+            my ($name, $coderef) = @_;
+
+            ### _make_delegation_method() called with a: ref $coderef
+            return $self->$orig(@_)
+                unless 'CODE' eq ref $coderef;
+
+            # this coderef will be installed as a method on the associated class itself.
+            my $custom_coderef = sub {
+                # aka $self from the class instance's perspective
+                my $associated_class_instance = shift @_;
+
+                # in $coderef, $_ will be the attribute metaclass
+                local $_ = $self;
+                return $associated_class_instance->$coderef(@_);
+            };
+
+            return $self->_process_accessors(custom => { $name => $custom_coderef });
+        };
+
+        return;
     };
 }
 
@@ -273,10 +367,10 @@ sub init_meta {
     die "Class $for_class has no metaclass!"
         unless Class::MOP::class_of($for_class);
 
-    # If we're given paramaters to pass on to construct a role with, we build
+    # If we're given parameters to pass on to construct a role with, we build
     # it out here rather than pass them on and allowing apply_metaroles() to
-    # handle it, as there are Very Loud Warnings about how paramatized roles
-    # are non-cachable when generated on the fly.
+    # handle it, as there are Very Loud Warnings about how parameterized roles
+    # are non-cacheable when generated on the fly.
 
     ### $params
     my $role
@@ -288,10 +382,12 @@ sub init_meta {
         ;
 
     Moose::Util::MetaRole::apply_metaroles(
-        for             => $for_class,
-        class_metaroles => { attribute         => [ $role ] },
         # TODO add attribute trait here to create builder method if found
-        role_metaroles  => { applied_attribute => [ $role ] },
+        for                          => $for_class,
+        class_metaroles              => { attribute         => [ $role ] },
+        role_metaroles               => { applied_attribute => [ $role ] },
+        parameter_metaroles          => { applied_attribute => [ $role ] },
+        parameterized_role_metaroles => { applied_attribute => [ $role ] },
     );
 
     return Class::MOP::class_of($for_class);
@@ -303,9 +399,12 @@ __END__
 
 =pod
 
-=encoding utf-8
+=encoding UTF-8
 
-=for :stopwords Chris Weyl GitHub attribute's isa one's rwp
+=for :stopwords Chris Weyl David Steinbrunner GitHub attribute's isa one's rwp SUBTYPING
+foo
+
+=for :stopwords Wishlist flattr flattr'ed gittip gittip'ed
 
 =head1 NAME
 
@@ -313,7 +412,7 @@ MooseX::AttributeShortcuts - Shorthand for common attribute options
 
 =head1 VERSION
 
-This document describes version 0.018 of MooseX::AttributeShortcuts - released January 09, 2013 as part of MooseX-AttributeShortcuts.
+This document describes version 0.025 of MooseX::AttributeShortcuts - released May 22, 2014 as part of MooseX-AttributeShortcuts.
 
 =head1 SYNOPSIS
 
@@ -508,9 +607,27 @@ e.g., in your class,
     has foo => (is => 'ro', builder => '_build_foo');
     sub _build_foo { 'bar!' }
 
+=head2 isa_instance_of => ...
+
+Given a package name, this option will create an C<isa> type constraint that
+requires the value of the attribute be an instance of the class (or a
+descendant class) given.  That is,
+
+    has foo => (is => 'ro', isa_instance_of => 'SomeThing');
+
+...is effectively the same as:
+
+    use Moose::TypeConstraints 'class_type';
+    has foo => (
+        is  => 'ro',
+        isa => class_type('SomeThing'),
+    );
+
+...but a touch less awkward.
+
 =head2 isa => ..., constraint => sub { ... }
 
-Specifying the constraint option with a coderef will cause a new type
+Specifying the constraint option with a coderef will cause a new subtype
 constraint to be created, with the parent type being the type specified in the
 C<isa> option and the constraint being the coderef supplied here.
 
@@ -546,10 +663,135 @@ C<thinger> will correctly coerce the string "/etc/passwd" to a
 C<Path::Class:File>, and will only accept the coerced result as a value if
 the file exists.
 
+=head2 coerce => [ Type => sub { ...coerce... }, ... ]
+
+Specifying the coerce option with a hashref will cause a new subtype to be
+created and used (just as with the constraint option, above), with the
+specified coercions added to the list.  In the passed hashref, the keys are
+Moose types (well, strings resolvable to Moose types), and the values are
+coderefs that will coerce a given type to our type.
+
+    has bar => (
+        is     => 'ro',
+        isa    => 'Str',
+        coerce => [
+            Int    => sub { "$_"                       },
+            Object => sub { 'An instance of ' . ref $_ },
+        ],
+    );
+
+=head2 handles => { foo => sub { ... }, ... }
+
+Creating a delegation with a coderef will now create a new, "custom accessor"
+for the attribute.  These coderefs will be installed and called as methods on
+the associated class (just as readers, writers, and other accessors are), and
+will have the attribute metaclass available in $_.  Anything the accessor
+is called with it will have access to in @_, just as you'd expect of a method.
+
+e.g., the following example creates an attribute named 'bar' with a standard
+reader accessor named 'bar' and two custom accessors named 'foo' and
+'foo_too'.
+
+    has bar => (
+
+        is      => 'ro',
+        isa     => 'Int',
+        handles => {
+
+            foo => sub {
+                my $self = shift @_;
+
+                return $_->get_value($self) + 1;
+            },
+
+            foo_too => sub {
+                my $self = shift @_;
+
+                return $self->bar + 1;
+            },
+        },
+    );
+
+...and later,
+
+Note that in this example both foo() and foo_too() do effectively the same
+thing: return the attribute's current value plus 1.  However, foo() accesses
+the attribute value directly through the metaclass, the pros and cons of
+which this author leaves as an exercise for the reader to determine.
+
+You may choose to use the installed accessors to get at the attribute's value,
+or use the direct metaclass access, your choice.
+
+=head1 ANONYMOUS SUBTYPING AND COERCION
+
+    "Abusus non tollit usum."
+
+Note that we create new, anonymous subtypes whenever the constraint or
+coercion options are specified in such a way that the Shortcuts trait (this
+one) is invoked.  It's fully supported to use both constraint and coerce
+options at the same time.
+
+This facility is intended to assist with the creation of one-off type
+constraints and coercions.  It is not possible to deliberately reuse the
+subtypes we create, and if you find yourself using a particular isa /
+constraint / coerce option triplet in more than one place you should really
+think about creating a type that you can reuse.  L<MooseX::Types> provides
+the facilities to easily do this, or even a simple L<constant> definition at
+the package level with an anonymous type stashed away for local use.
+
+=head1 SEE ALSO
+
+Please see those modules/websites for more information related to this module.
+
+=over 4
+
+=item *
+
+L<MooseX::Types|MooseX::Types>
+
+=back
+
+=head1 SOURCE
+
+The development version is on github at L<http://https://github.com/RsrchBoy/moosex-attributeshortcuts>
+and may be cloned from L<git://https://github.com/RsrchBoy/moosex-attributeshortcuts.git>
+
+=head1 BUGS
+
+Please report any bugs or feature requests on the bugtracker website
+https://github.com/RsrchBoy/moosex-attributeshortcuts/issues
+
+When submitting a bug or request, please include a test-file or a
+patch to an existing test-file that illustrates the bug or desired
+feature.
+
 =head1 AUTHOR
 
 Chris Weyl <cweyl@alumni.drew.edu>
 
+=head2 I'm a material boy in a material world
+
+=begin html
+
+<a href="https://www.gittip.com/RsrchBoy/"><img src="https://raw.githubusercontent.com/gittip/www.gittip.com/master/www/assets/%25version/logo.png" /></a>
+<a href="http://bit.ly/rsrchboys-wishlist"><img src="http://wps.io/wp-content/uploads/2014/05/amazon_wishlist.resized.png" /></a>
+<a href="https://flattr.com/submit/auto?user_id=RsrchBoy&url=https%3A%2F%2Fgithub.com%2FRsrchBoy%2Fmoosex-attributeshortcuts&title=RsrchBoy's%20CPAN%20MooseX-AttributeShortcuts&tags=%22RsrchBoy's%20MooseX-AttributeShortcuts%20in%20the%20CPAN%22"><img src="http://api.flattr.com/button/flattr-badge-large.png" /></a>
+
+=end html
+
+Please note B<I do not expect to be gittip'ed or flattr'ed for this work>,
+rather B<it is simply a very pleasant surprise>. I largely create and release
+works like this because I need them or I find it enjoyable; however, don't let
+that stop you if you feel like it ;)
+
+L<Flattr this|https://flattr.com/submit/auto?user_id=RsrchBoy&url=https%3A%2F%2Fgithub.com%2FRsrchBoy%2Fmoosex-attributeshortcuts&title=RsrchBoy's%20CPAN%20MooseX-AttributeShortcuts&tags=%22RsrchBoy's%20MooseX-AttributeShortcuts%20in%20the%20CPAN%22>,
+L<gittip me|https://www.gittip.com/RsrchBoy/>, or indulge my
+L<Amazon Wishlist|http://bit.ly/rsrchboys-wishlist>...  If you so desire.
+
+=head1 CONTRIBUTOR
+
+David Steinbrunner <dsteinbrunner@pobox.com>
+
 =head1 COPYRIGHT AND LICENSE
 
 This software is Copyright (c) 2011 by Chris Weyl.
